from .command_runner import program_exists


def ffmpeg_warning() -> str:
    """Print a warning if FFMPEG is not found in path"""
    if not program_exists("ffmpeg"):
        return "[bold red]~~~WARNING~~~ FFMPEG is not found in the system path. Many functions will not work without it."
    return ""


class HelpText:
    """Contains the description of commands and parameters for each of the user facing functions
    Error text is handled within individual functions"""

    # General info
    license = "[blue bold link=https://https://github.com/jonathanfox5/gogadget]gogadget[/] © Jonathan Fox is licensed under [blue bold link=https://creativecommons.org/licenses/by-nc-sa/4.0/?ref=chooser-v1]CC BY-NC-SA 4.0[/]."

    # Commands
    main = """
    gogadget is a toolkit for producing immersion and priming tools for language learning. It is capable of downloading audio and video files, automatically transcribing subtitles from videos and podcasts, and automatically producing filtered Anki decks with sentence audio / translations / screenshots / definitions.

    [bright_white i][u]Basics[/u]
    All commands are listed in the "Primary Functions" box below and have their own documentation. Each command has parameters associated with it. These can be listed by just typing [b magenta]gogadget[/] then the name of the command that you are interested in. For example:
    [b magenta]gogadget download[/]

    You will see from the output of that command that you can just run the following to download a video:
    [b magenta]gogadget download --url "https://www.videosite.com/watch?v=videoid"[/]

    [u]Advanced[/u]
    Commands have both a "standard" form and a "short" form. You can use whatever works best for you! The following two lines are equivalent.
    [b magenta]gogadget download --url "https://www.videosite.com/watch?v=videoid" --output "immersion videos" --subtitle_language en
    gogadget download -i "https://www.videosite.com/watch?v=videoid" -o "immersion videos" -l en[/]
    
    Note: Regardless of the "standard" name, all commands follow the same logic for their "short" names. The item that is being used as input is -i, the output is -o and the language is -l. Normally you don't need any more than this!

    To get a list of supported languages and the associated two letter codes, run this command:
    [b magenta]gogadget list-languages[/]

    [u]Configuration[/u]
    It's recommended, but not required, that you fully install the models for the languages that you are interested in. e.g. To install Italian (target language) with English (native language) translations, run:
    [b magenta]gogadget install --language it --translation-language en[/]

    You can also configure defaults so that you don't need to specify as many parameters each time you run your commands:
    [b magenta]gogadget set-defaults --custom[/][/]
    """
    install = """Download models for a given --language and initialises tools.

    [bright_white i][u]Examples:[/u]
    1. Install modules to process Italian and produce English translations.
    gogadget install --language it --translation-language en

    2. To get a list of language codes to use in the command, run:
    gogadget list-languages
    """
    set_defaults = """
    Configure your default paths so that don't need to specify them each time.
    
    [bright_white i][u]Examples:[/u]

    1. Open the settings file on your folder in your default text editor.
    gogadget set-defaults --custom

    2. Reset to factory defaults.
    gogadget set-defaults --factory

    [red b]~~~~ WARNING ~~~~ It is possible to break the tool by setting incorrect values in the config file.
    Reset to factory defaults if you experience errors or unexpected behaviour.
    """
    update_downloader = """
    Update the downloader to use the latest version of yt-dlp.

    [bright_white i][u]Examples:[/u]

    1. Update downloader.
    gogadget update-downloader
    """
    list_languages = """
    Display languages supported by the tool.

    [bright_white i][u]Examples:[/u]
    1. List languages supported by all functions of the tool.
    gogadget list-languages

    2. List languages supported or partially supported by each module.
    gogadget list-languages --detailed
    """
    download = """
    Download a video or playlist from a website URL. 
    
    [bright_white i][u]Examples:[/u]

    1. Normal usage using standard names.
    gogadget download --url "https://www.videosite.com/watch?v=videoid"
    
    2. More advanced usage using standard names.
    gogadget download --url "https://www.videosite.com/watch?v=videoid" --output "immersion videos" --subtitle_language en --format "best"

    3. Equivalent of (2) using short names.
    gogadget download -i "https://www.videosite.com/watch?v=videoid" -o "immersion videos" -l en -f "best"
    """
    download_audio = """
    Download a video or playlist from a website URL and convert it to an audio file.
    
    [bright_white i][u]Examples:[/u]

    1. Normal usage using standard names.
    gogadget download-audio --url "https://www.videosite.com/watch?v=videoid"
    
    2. More advanced usage using standard names.
    gogadget download-audio  --url "https://www.videosite.com/watch?v=videoid" --output "immersion videos"

    3. Equivalent of (2) using short names.
    gogadget download-audio  -i "https://www.videosite.com/watch?v=videoid" -o "immersion videos"
    """
    download_subtitles = """
    Download subtitles from an online video service.

    [bright_white i][u]Examples:[/u]
    1. Download english subtitles for a given video.
    gogadget download-subtitles --url "https://www.videosite.com/watch?v=videoid" --subtitle-language en

    2. Equivalent of (1) using short names.
    gogadget download-subtitles -i "https://www.videosite.com/watch?v=videoid" -l en
    """
    transcribe = """
    Produce subtitle file(s) from audio or video using whisperX.

    [bright_white]--input and -i accept both files and directories of files.

    If you have an NVIDIA GPU that is set up for CUDA, it's strongly recommended to pass the --gpu flag as this significantly speeds up the tool.

    You can also reduce runtime (at the expense of accuracy) by specifying --whisper-model small

    [i][u]Examples:[/u]
    1. Transcribe a media file or folder of media files that is in English.
    gogadget transcribe --input "path to media file or folder containing media files" --language en

    2. As per (1) but using the GPU to process the model.
    gogadget transcribe --input "path to media file or folder containing media files" --language en --gpu

    3. Example using short names where the output folder is also specified.
    gogadget transcribe -i "path to media file or folder containing media files" -o "folder to save to" -l en -g
    """
    create_anki_deck = """
    Build an Anki deck using the most common vocabulary in a subtitles file or a folder of subtitles. Optionally include audio and / or screenshots from the source media file(s).
    
    [bright_white]If you use this regularly, it's highly recommended to set the default paths to your dictionary, excluded words, etc. and preferred processing options to simplify the process.
    You can set your defaults using the following command: 
    [i]gogadget set-defaults --custom

    [u]Examples:[/u]
    1. Normal usage using standard names where your target language is italian and your native language is English.
    gogadget anki-deck --input "folder containing subtitles and media files" --language it --translation-language en

    2. As per (1) but uses dictionary, word exclude list and word audio bank. Also uses --exclude-no-definition to filter out proper nouns / non-target language words.
    gogadget anki-deck --input "folder containing subtitles and media files" --language it --translation-language en --dictionary "dictionary.json" --word_audio "folder_name" --excluded-words "excel_name.xlsx" --exclude-no-definition
    
    3. Equivalent of (2) using short names.
    gogadget anki-deck -i "folder containing subtitles and media files" -l it -t en -d "dictionary.json" -w "folder_name" -e "excel_name.xlsx" -h
    
    4. If you have set all of your defaults as described above, you can just run.
    gogadget anki-deck -i "folder containing subtitles and media files"
    """
    frequency_analysis = """
    Produce a frequency analysis of the most common vocabulary in a subtitles file or a folder of subtitles. Useful for priming, also used as a pre-processing stage for some other functions.
    
    [bright_white]If you use this regularly, it's highly recommended to set the default paths to your excluded words and preferred processing options to simplify the process.
    You can set your defaults using the following command: 
    [i]gogadget set-defaults --custom

    [u]Examples:[/u]
    1. Normal usage using standard names where your target language is italian.
    gogadget frequency-analysis --input "folder containing subtitles and media files" --language it

    2. As per (1) but uses word exclude list.
    gogadget frequency-analysis --input "folder containing subtitles and media files" --language it --excluded-words "excel_name.xlsx"
    
    3. Equivalent of (2) using short names.
    gogadget frequency-analysis -i "folder containing subtitles and media files" -l it -e "excel_name.xlsx"

    4. If you have set all of your defaults as described above, you can just run.
    gogadget frequency-analysis -i "folder containing subtitles and media files"
    """
    interactive_transcript = "Create an interactive transcript from a video or audio source."

    # Parameters
    video_url = "URL of the video or playlist. Supports any website supported by [blue underline link=https://github.com/yt-dlp/yt-dlp/blob/master/supportedsites.md]yt-dlp[/]."
    language_code = "Language to use for processing. This should be a two letter language code, e.g. [cyan b]en[/] (for English), [cyan b]es[/] (for Spanish) or [cyan b]it[/] (Italian). Run [cyan bold]gogadget list-languages[/] for a list of supported languages."
    filter_sheet_path = "[cyan][Optional][/] Path to spreadsheet which has list of words to filter. Words need to be in column 'A' but can be entered across multiple sheets."
    application_version = "Display application version."
    video_format = "[cyan][Optional][/] Specify the format of the video. Accepts [blue underline link=https://github.com/yt-dlp/yt-dlp?tab=readme-ov-file#format-selection]yt-dlp's format options[/]."
    output_directory = "[cyan][Optional][/] Directory (aka folder) to save the files to. Defaults to the current working directory where the user is running the script from."
    ytdlp_options = "[cyan][Optional][/][red][Advanced][/] Custom yt-dlp options, should accept any command line arguments on the [blue underline link=https://github.com/yt-dlp/yt-dlp]github page[/]. Please format these as a string, enclosed by quotes."
    ytdlp_subtitles = "[cyan][Optional][/] Language of subtitles to download. If you want to download these, you should enter a two letter language code such as [cyan b]en[/], [cyan b]es[/] or [cyan b]it[/]. It will try to download manual subtitles first and fallback to automatically generated subtitles if these aren't found."
    ytdlp_subtitles_required = "Language of subtitles to download. You should enter a two letter language code such as [cyan b]en[/], [cyan b]es[/] or [cyan b]it[/]. It will try to download manual subtitles first and fallback to automatically generated subtitles if these aren't found."
    dictionary_file = "[cyan][Optional][/] Dictionary in json format to retrieve definitions from for the Anki cards."
    word_examples = "[cyan][Optional][/] Directory of mp3 files of individual words to include in the Anki cards."
    transcribe_path = "Path to the video or audio file to transcribe. This can be either a specific video / audio file or a folder of files."
    whisper_model = "[cyan][Optional][/] Specify the whisper model to use for transcription. By default, this is large-v3 turbo but setting this to [cyan b]small[/] can significantly speed the process up at the cost of accuracy."
    alignment_model = "[cyan][Optional][/] Specify the model from hugging face to use to align the subtitles with the audio. For the most common languages, the tool will find this for you."
    use_cuda = "[cyan][Optional][/] You can specify --gpu if you have a CUDA enabled Nvidia graphics card to significantly speed up the processing."
    subtitle_format = "[cyan][Optional][/] File format for the subtitles. You can specify vtt, srt, json, txt, tsv or aud. Vtt is the preferred format of the other tools in this suite."
    lemmatise = "[cyan][Optional][/] Enable or disable lemmatisation. If supported for your language, this is generally recommended."
    exclude_sheet = "[cyan][Optional][/] Spreadsheet containing words to exclude from the analysis (e.g. the most common words in a language, words already learned). Words should be in the first column of the spreadsheet but can be split across multiple sub-sheets within the file."
    input_directory_anki = "Directory (folder) containing the video file(s) and subtitle files(s) to be turned into an Anki deck."
    subtitle_offset = "[cyan][Optional][/] Time, in milliseconds, to offset the subtitles by when extracting audio. Not normally required if subtitles were generated by [cyan bold]gogadget transcribe[/]."
    subtitle_buffer = "[cyan][Optional][/] Extra time, in milliseconds, to add to the extracted audio to avoid it being cut off. Not normally required if subtitles were generated by [cyan bold]gogadget transcribe[/]."
    max_cards = "[cyan][Optional][/] Maximum number of cards to include in the deck."
    anki_media = "[cyan][Optional][/] Media to extract sentence audio and screenshots from to display on the Anki card. This can either be a video or audio only source."
    include_no_definition = "[cyan][Optional][/] Include cards where the definition can't be found in the dictionary. Setting --exclude-no-definition may improve the quality of the deck as it will likely filter many proper nouns, words not from the target language, etc."
    stop_words = "[cyan][Optional][/] If lemmatisation is enabled, you can include or exclude stop words. Stop words are short 'function' words such as 'the', 'that', 'which', etc."
    language_code_translation = "[cyan][Optional][/] Language to use for translations. Translation quality is generally best if either the target language or the translation is set to [cyan b]en[/] (English). This should be a two letter language code, e.g. [cyan b]en[/] (for English), [cyan b]es[/] (for Spanish) or [cyan b]it[/] (Italian). Run [cyan bold]gogadget list-languages[/] for a list of supported languages."
    defaults_factory = "[cyan][Optional][/] Load factory default settings. These settings are chosen to be compatible with most systems and languages with minimal tweaking."
    defaults_optimised = "[cyan][Optional][/][red][Advanced][/] Load 'optimised' settings. Results are generally higher quality and tool execution time is reduced. However, some tweaking may be required to get this to run with some systems and language. Generally used in combination with [cyan bold]gogadget set-defaults --custom[/]."
    defaults_custom = "[cyan][Optional][/] Set custom settings in a text file. Useful for setting default paths to resources."
    list_languages_detailed = (
        "[cyan][Optional][/] List the languages supported by each module of the tool."
    )
    install_gpu = "[cyan][Optional][/][red][Advanced][/] Install CUDA for an NVIDIA gpu to speed up transcription."

    # .toml file
    toml_instructions = """IMPORTANT INFORMATION
# =====================
# - All values are text and should be therefore be wrapped in double quotes. Valid examples:
#       language = "en"
#       lemmatise = "True"
#       lemmatise = "False"
#       subs_offset_ms = "0"
#       subs_offset_ms = "50"
# - If you don't want to a specify a value, just type two double quotes beside each other e.g.:
#       language = ""
#       word_exclude_spreadsheet = ""
# - If you are on Windows, any paths will need to have any backslashes replaces with a double backslash e.g.:
#       word_exclude_spreadsheet = "C:\\\\data\\\\exclude.xlsx"
#   Since this is easy to forget about, the tool will try to fix it for you. However, it's always best if it is correct to begin with!
#
# WARNING
# =======
# It is possible to break the tool by setting incorrect values in here.
# However, the script will attempt to fall back to sensible defaults if it can't read your values.
# If your setting appears to not be read by the tool, this is probably the reason!
# Run `gogadget set-defaults --factory` (without quotes) to reset this file if you run into errors or unexplained behaviour"""
    toml_general = """language and language_for_translations either be a valid two letter language code or be set to "". 
# Valid examples:
#       language = "en"
#       language = ""
# For a list of supported languages, please see the readme or run `gogadget list-languages` (without quotes)
#
# output_directory needs to be a valid folder on your system.
# You can use a dot "." if you want to use the current directory that you are running commands from.
# Windows paths need to have backslashes replaced with double backslashes, see [instructions] at the top of this file.
# The tool will try to fix it if you forget but it's best to get it correct to begin with!
# Valid examples:
#       output_directory = ""                         # No default, you will have to specify when running the command
#       output_directory = "."                        # The outputs of the command will be written to the current folder
#       output_directory = "immersion_videos"         # Outputs will be written to a sub folder called "immersion_videos"
#       output_directory = "C:\\\\immersion_videos\\\\"   # Outputs will be written to a specific folder on the C:\\ drive
"""
    toml_external = """These can be set to "" if you don't want to use them or want to specify them every time.
# Windows paths need to have backslashes replaced with double backslashes, see [instructions] at the top of this file.
# The tool will try to fix it if you forget but it's best to get it correct to begin with!
# Valid examples:
#       word_exclude_spreadsheet = "C:\\\\data\\\\exclude.xlsx"     # This will load a specific spreadsheet
#       word_exclude_spreadsheet = ""                           # Don't use an exclude spreadsheet or only use when specified in the command
"""
    toml_anki = """extract_media and include_words_with_no_definition should either be set to "True" or "False" and MUST be wrapped in quotes. 
# Valid examples:
#       extract_media = "True"
#       include_words_with_no_definition = "False"
#
# subs_offset_ms, subs_buffer_ms and max_cards_in_deck should be a number wrapped in quotes. 
# Valid examples:
#       subs_offset_ms = "0"
#       subs_buffer_ms = "50
"""
    toml_lemmatiser = """All values should be set to "True" or "False" and MUST be wrapped in quotes.
# Valid examples:
#       lemmatise = "True"
#       lemmatise = "False"
"""
    toml_downloader = """These should either wrapped in quotes or set to double quotes to leave it blank. 
# Valid examples:
#       format = "best[ext=mp4]"
#       format = ""
"""
    toml_transcriber = """whisper_use_gpu should either be set to "True" or "False" and MUST be wrapped in quotes. 
# Valid examples:
#       whisper_use_gpu = "False"
#       whisper_use_gpu = "True"
#
# The other settings should be text wrapped in quotes or be set to "" if you want to specify them each time.
# These settings are best left alone unless you know what you are doing! Valid examples:
#       whisper_model = "small"
#       alignment_model = ""
"""


class SupportedLanguages:
    """Contains dictionaries of the supported languages for each function plus some explanation text"""

    common_explanation = "All features of the tool are available for the following languages. [b]If your language isn't listed[/], you can run [cyan bold]gogadget list-languages --detailed[/] to get a list of features that are supported by each language."
    whisper_explanation = "The transcriber (whisperX) supports the following languages fully. Languages not listed might work or might work with some manual configuration. However, basic functionality is available for over 100 languages! The transcriber is not required if you provide your own subtitle files."
    spacy_explanation = "The lemmatiser (spacy) supports the following languages. The tool will still work if your language is not available. However, the tool will not be able to lemmatise words or remove 'stop' words."
    argos_explanation = "The translator (argos) supports the following languages. The tool will still work if your language is not available. However, the tool will not be able to translate sentences."
    whisper_languages = {
        "ar": "Arabic",
        "ca": "Catalan",
        "zh": "Chinese",
        "hr": "Croatian",
        "cs": "Czech",
        "da": "Danish",
        "nl": "Dutch",
        "en": "English",
        "fi": "Finnish",
        "fr": "French",
        "de": "German",
        "he": "Hebrew",
        "hi": "Hindi",
        "hu": "Hungarian",
        "it": "Italian",
        "ja": "Japanese",
        "ko": "Korean",
        "ml": "Malayalam",
        "el": "Greek",
        "no": "Norwegian",
        "nn": "Norwegian Nynorsk",
        "fa": "Persian",
        "pl": "Polish",
        "pt": "Portuguese",
        "ru": "Russian",
        "sk": "Slovak",
        "sl": "Slovenian",
        "es": "Spanish",
        "te": "Telugu",
        "tr": "Turkish",
        "uk": "Ukrainian",
        "ur": "Urdu",
        "vi": "Vietnamese",
    }

    spacy_languages = {
        "ca": "Catalan",
        "zh": "Chinese",
        "hr": "Croatian",
        "da": "Danish",
        "nl": "Dutch",
        "en": "English",
        "fi": "Finnish",
        "fr": "French",
        "de": "German",
        "el": "Greek",
        "it": "Italian",
        "ja": "Japanese",
        "ko": "Korean",
        "lt": "Lithuanian",
        "mk": "Macedonian",
        "nb": "Norwegian Bokmål",
        "pl": "Polish",
        "pt": "Portuguese",
        "ro": "Romanian",
        "ru": "Russian",
        "sl": "Slovenian",
        "es": "Spanish",
        "sv": "Swedish",
        "uk": "Ukrainian",
    }

    argos_languages = {
        "sq": "Albanian",
        "ar": "Arabic",
        "az": "Azerbaijani",
        "eu": "Basque",
        "bn": "Bengali",
        "bg": "Bulgarian",
        "ca": "Catalan",
        "zh": "Chinese",
        "zt": "Chinese (traditional)",
        "cs": "Czech",
        "da": "Danish",
        "nl": "Dutch",
        "en": "English",
        "eo": "Esperanto",
        "et": "Estonian",
        "fi": "Finnish",
        "fr": "French",
        "gl": "Galician",
        "de": "German",
        "el": "Greek",
        "he": "Hebrew",
        "hi": "Hindi",
        "hu": "Hungarian",
        "id": "Indonesian",
        "ga": "Irish",
        "it": "Italian",
        "ja": "Japanese",
        "ko": "Korean",
        "lv": "Latvian",
        "lt": "Lithuanian",
        "ms": "Malay",
        "nb": "Norwegian",
        "fa": "Persian",
        "pl": "Polish",
        "pt": "Portuguese",
        "ro": "Romanian",
        "ru": "Russian",
        "sk": "Slovak",
        "sl": "Slovenian",
        "es": "Spanish",
        "sv": "Swedish",
        "tl": "Tagalog",
        "th": "Thai",
        "tr": "Turkish",
        "uk": "Ukrainian",
        "ur": "Urdu",
    }

    @staticmethod
    def get_common_languages():
        """Works out the languages that are supported by all functions by combining dictionaries"""

        a = SupportedLanguages.spacy_languages
        b = SupportedLanguages.whisper_languages
        c = SupportedLanguages.argos_languages

        output_dict = {}

        for key, value in a.items():
            if (key in b) and (key in c):
                output_dict[key] = value

        return output_dict



import re
from pathlib import Path

import pandas as pd
import pysubs2
from lemon_tizer import LemonTizer
from rich.progress import track

from .cli_utils import CliUtils
from .config import SUPPORTED_SUB_EXTS, ConfigFile
from .lemmatiser import language_supported
from .utils import (
    dataframe_to_excel,
    generate_output_path,
    import_first_column_from_sheets,
    list_files_with_extension,
)


def frequency_analyser(
    input_path: Path,
    output_directory: Path,
    language: str,
    lemmatise_words: bool,
    filter_stop_words: bool,
    exclude_spreadsheet_path: Path | None = None,
) -> pd.DataFrame:
    """Main entry point for the frequency analyser function"""

    # Sort out paths
    sub_sheet_save_path = generate_output_path(input_path, output_directory, "subs", "xlsx")
    frequency_analysis_save_path = generate_output_path(
        input_path, output_directory, "frequency_analysis", "xlsx"
    )

    # Arrange the subs in a dataframe
    df_subs = subs_to_dataframe(input_path, sub_sheet_save_path)
    if df_subs is None:
        return None

    # Disable lemmatisation if spacy doesn't support it
    if lemmatise_words and (not language_supported(language)):
        lemmatise_words = False

        CliUtils.print_warning(
            f"Language {language} is not currently supported by the lemmatiser so lemmatisation has been disabled. However, the script will still function. Without lemmatisation, it is highly recommended to set --excluded-words to a spreadsheet of common words for {language} for functions that support it. Type [cyan bold]gogadget [command_name_here] --help[/] for more information for each individual command"
        )

    # Get the list of words to filter
    df_exclude = None
    if exclude_spreadsheet_path is not None:
        if (not exclude_spreadsheet_path.is_file()) or (not exclude_spreadsheet_path.exists()):
            CliUtils.print_warning(
                f"{exclude_spreadsheet_path} is not an existing spreadsheet file. Tool will proceed but words will not be filtered."
            )
        else:
            df_exclude = import_first_column_from_sheets(exclude_spreadsheet_path)

    # Do the frequency analysis
    df_freq = generate_frequency_analysis(
        df_input=df_subs,
        language=language,
        lemmatise_words=lemmatise_words,
        df_exclude=df_exclude,
        frequency_analysis_save_path=frequency_analysis_save_path,
        filter_stop_words=filter_stop_words,
    )

    return df_freq


def filter_frequency_analysis(
    df_main: pd.DataFrame,
    df_exclude: pd.DataFrame,
    field_main: str = "word",
    field_exclude: str = "word",
) -> pd.DataFrame:
    """Remove excluded words from the analysis output"""
    df_main = df_main[~df_main[field_main].isin(df_exclude[field_exclude])]

    return df_main


def generate_frequency_analysis(
    df_input: pd.DataFrame,
    language: str,
    lemmatise_words: bool,
    filter_stop_words: bool,
    df_exclude: pd.DataFrame | None = None,
    frequency_analysis_save_path: Path | None = None,
) -> pd.DataFrame:
    """Main helper function to do the frequency analysis"""

    # TODO: [V2] Generalise so that it will work with things that aren't subs

    CliUtils.print_status("Frequency analysis: loading language model")

    CONFIG = ConfigFile()
    CONFIG.read_defaults()

    filter_punctuation = CONFIG.lemmatiser.filter_out_non_alpha
    if lemmatise_words:
        lt = LemonTizer(language=language, model_size="lg")
        lt.set_lemma_settings(
            filter_out_non_alpha=filter_punctuation,
            filter_out_common=CONFIG.lemmatiser.filter_out_stop_words,
            convert_input_to_lower=CONFIG.lemmatiser.convert_input_to_lower,
            convert_output_to_lower=CONFIG.lemmatiser.convert_output_to_lower,
            return_just_first_word_of_lemma=CONFIG.lemmatiser.return_just_first_word_of_lemma,
        )

    CliUtils.print_status("Frequency analysis: counting words")
    df_freq_table = pd.DataFrame(
        columns=["word", "example", "example_source", "example_start", "example_end", "frequency"]
    )

    for index, row in track(
        df_input.iterrows(),
        description="Frequency analysis: counting words...",
        total=len(df_input),
    ):
        line = row["text"]
        example_source = row["filename"]
        sentence_start = row["start"]
        sentence_end = row["end"]

        if lemmatise_words:
            word_list = lt.lemmatize_sentence(input_str=line)
        else:
            word_list = split_sentence(input_str=line, filter_punctuation=filter_punctuation)

        for word_dict in word_list:
            for word, word_lemma in word_dict.items():
                word_key = word_lemma

                if word_key in df_freq_table["word"].values:
                    df_freq_table.loc[df_freq_table["word"] == word_key, "frequency"] += 1
                else:
                    new_row = pd.DataFrame(
                        [
                            {
                                "word": word_key,
                                "example": line,
                                "example_source": example_source,
                                "example_start": sentence_start,
                                "example_end": sentence_end,
                                "frequency": 1,
                            }
                        ]
                    )
                    df_freq_table = pd.concat([df_freq_table, new_row], ignore_index=True)

    CliUtils.print_status("Frequency analysis: post-processing")

    # Filter out
    if df_exclude is not None:
        df_freq_table = filter_frequency_analysis(df_freq_table, df_exclude)

    # Sort and export
    df_freq_table = df_freq_table.sort_values(by=["frequency"], ascending=False)

    if frequency_analysis_save_path is not None:
        dataframe_to_excel(df_freq_table, frequency_analysis_save_path)

    return df_freq_table


def split_sentence(input_str: str, filter_punctuation: bool) -> list[dict[str, str]]:
    """Split sentence into words. Used if lemmatiser isn't being used.
    Objective of this is to mirror the output of the lemmatiser for languages that aren't supported
    """

    if filter_punctuation:
        input_str = remove_punctuation(input_str)

    input_str = input_str.lower()
    input_list = input_str.split(" ")

    result: list[dict[str, str]] = []
    for word in input_list:
        processed_word = word.strip()
        result.append({processed_word: processed_word})
    return result


def remove_punctuation(input_string: str) -> str:
    """Remove punctuation from a string. Used if lemmatiser isn't being used.
    Only allow letters/numbers, hyphens, apostrophes and whitespace
    Untested on languages that don't use the latin alphabet but google seems to claim it works...
    """

    pattern = r"[^\w\s'-]"

    output_string = re.sub(pattern, "", input_string)

    return output_string


def subs_to_dataframe(
    input_path: Path,
    spreadsheet_output_path: Path | None = None,
) -> pd.DataFrame:
    """Pull subtitle data from files into a dataframe"""

    path_list = list_files_with_extension(
        input_path, valid_suffixes=SUPPORTED_SUB_EXTS, file_description_text="subtitle"
    )
    if len(path_list) == 0:
        return None

    df_subs = _extract_subs(path_list)

    if spreadsheet_output_path is not None:
        dataframe_to_excel(df_subs, spreadsheet_output_path)

    return df_subs


def _extract_subs(path_list: list[Path]) -> pd.DataFrame:
    """Helper function for subs_to_dataframe"""

    df = pd.DataFrame(columns=["filename", "start", "end", "text"])
    for path in path_list:
        filename = str(path)

        subs = pysubs2.load(path.resolve())
        for sub in subs:
            add_row = pd.DataFrame(
                [{"filename": filename, "start": sub.start, "end": sub.end, "text": sub.text}]
            )
            df = pd.concat([df, add_row], ignore_index=True)

    return df



import warnings

import argos_spacy_compatibility.package as pkg
import argos_spacy_compatibility.translate as argos
import pandas as pd
from argos_spacy_compatibility.package import Package as pkg_type
from rich.progress import track

from .cli_utils import CliUtils


def translate_all_sentences(
    df: pd.DataFrame, from_lang: str, to_lang: str, source_column: str = "example"
) -> pd.DataFrame:
    """Translate all sentences in a dataframe. Write the translations to the dataframe."""

    # Suppress stanza warnings so that we can see more important output in the console
    warnings.filterwarnings("ignore", category=FutureWarning, module="stanza")

    # Check we have the required languages installed. If not, install them.
    langs_installed = install(from_lang, to_lang)

    if not langs_installed:
        CliUtils.print_warning(
            f"Translator: could not load models for translating from {from_lang} to {to_lang}"
        )
        return df

    for index, row in track(df.iterrows(), description="Translating sentences...", total=len(df)):
        sentence = str(row[source_column])
        translation = translate_sentence(sentence, from_lang, to_lang)
        df.at[index, "translation"] = translation

    return df


def translate_sentence(sentence: str, from_lang: str, to_lang: str) -> str:
    """Translate an individual sentence"""

    warnings.filterwarnings("ignore", category=FutureWarning, module="stanza")
    try:
        translation = argos.translate(sentence, from_lang, to_lang)
    except Exception as e:
        CliUtils.print_error(f"Could not translate sentence from {from_lang} to {to_lang}", e)
        translation = ""

    return translation


def install(from_lang: str, to_lang: str) -> bool:
    """Work out the translation route between languages and install the appropriate argos packages, if required"""

    # Check if the requires languages have been installed
    CliUtils.print_status("Translator: checking installed languages")

    if from_lang == to_lang:
        CliUtils.print_warning(
            f"""Translator: target language and native language are both set to '{from_lang}' so no translation will be undertaken. 
    Try specifying --translation-language [your-native-language-2-letter-code-here] next time you run the command.
    For more information, run gogadget [command_name] --help"""
        )
        return False

    installed_packages = pkg.get_installed_packages()
    required_packages = translation_route(installed_packages, from_lang, to_lang)

    # If not, try to install them
    if len(required_packages) == 0:
        # Try to work out the translation route from the package index
        CliUtils.print_status("Translator: attempting language installation")
        pkg.update_package_index()
        available_packages = pkg.get_available_packages()
        required_packages = translation_route(available_packages, from_lang, to_lang)

        # Exit with failure message if we can't find a translation route
        if len(required_packages) == 0:
            CliUtils.print_error(
                f"Could not find a translation route from {from_lang} to {to_lang}"
            )
            return False

        # If we do have a translation route, install the languages
        install_languages(required_packages)

    CliUtils.print_rich("Translation route achieved:")
    CliUtils.print_rich(required_packages)

    return True


def install_languages(required_packages: list[pkg_type]):
    """Install each argos package in the list"""

    for package in required_packages:
        CliUtils.print_rich(f"Installing {str(package)}")

        pkg.install_from_path(package.download())


def translation_route(available_packages: list[pkg_type], from_lang: str, to_lang: str) -> list:
    """Work out the route that needs to be taken to translate between languages based upon a provided list of packages"""

    # Try to find a direct translation route first
    required_packages = []
    for package in available_packages:
        if (package.from_code == from_lang) and (package.to_code == to_lang):
            required_packages.append(package)
            return required_packages

    # If not available, try going via English
    found_to_en = False
    found_from_en = False

    for package in available_packages:
        if (package.from_code == from_lang) and (package.to_code == "en"):
            required_packages.append(package)
            found_to_en = True
            break

    for package in available_packages:
        if (package.from_code == "en") and (package.to_code == to_lang):
            required_packages.append(package)
            found_from_en = True
            break

    if (not found_to_en) or (not found_from_en):
        return []

    return required_packages



from pathlib import Path

import ffmpeg
import pandas as pd
from rich.progress import track

from .cli_utils import CliUtils
from .config import SUPPORTED_AUDIO_EXTS, SUPPORTED_SUB_EXTS, SUPPORTED_VIDEO_EXTS
from .utils import generate_output_path, list_files_with_extension


def extract_sentence_media(
    audio_dir: Path,
    screenshot_dir: Path,
    df: pd.DataFrame,
    subs_to_media: dict[Path, Path],
    subs_offset: int,
    subs_buffer: int,
) -> pd.DataFrame:
    """Extract screenshots and audio clips from media as specified by each row in a dataframe.
    Minimum required fields in df: example_source [str], example_start [int in ms], example_end [int in ms].
    DF should ideally also has blank fields for screenshot_path [str] and sentence_audio_path [str]
    """

    # Loop through and extract the media
    extract_log: list[str] = []
    for index, row in track(df.iterrows(), description="Extracting media...", total=len(df)):
        sub_path = Path(str(row["example_source"]))
        media_path = subs_to_media[sub_path]

        start_time = int(row["example_start"])
        end_time = int(row["example_end"])
        mid_time = int((start_time + end_time) / 2)
        buffered_start = max((start_time + subs_offset - subs_buffer), 1)
        buffered_end = end_time + subs_offset + subs_buffer

        # If the identified file is a video, grab the screenshot half way between the start and end time
        if media_path.suffix in SUPPORTED_VIDEO_EXTS:
            screenshot_path = extract_frame(
                video_path=media_path,
                output_directory=screenshot_dir,
                timestamp_ms=mid_time,
                extract_log=extract_log,
            )
            df.at[index, "screenshot_path"] = str(screenshot_path.resolve())
            extract_log.append(str(screenshot_path))

        # Assuming we aren't going to blow up ffmpeg, extract the the sentence audio
        if (media_path.suffix in SUPPORTED_VIDEO_EXTS) or (
            media_path.suffix in SUPPORTED_AUDIO_EXTS
        ):
            audio_path = extract_audio_segment(
                video_path=media_path,
                output_directory=audio_dir,
                start_time_ms=buffered_start,
                end_time_ms=buffered_end,
                extract_log=extract_log,
            )
            df.at[index, "sentence_audio_path"] = str(audio_path.resolve())
            extract_log.append(str(audio_path))

    return df


def match_subtitles_to_media(input_directory: Path) -> dict[Path, Path]:
    """Based upon file name, try to match subtitles with media files.
    If both video and audio have the same file name, prioritise video"""

    result: dict[Path, Path] = {}

    all_subs: list[Path] = list_files_with_extension(
        input_path=input_directory,
        valid_suffixes=SUPPORTED_SUB_EXTS,
        file_description_text="subtitles",
        search_subdirectories=False,
        print_errors=False,
    )
    all_videos: list[Path] = list_files_with_extension(
        input_path=input_directory,
        valid_suffixes=SUPPORTED_VIDEO_EXTS,
        file_description_text="videos",
        search_subdirectories=False,
        print_errors=False,
    )
    all_audio: list[Path] = list_files_with_extension(
        input_path=input_directory,
        valid_suffixes=SUPPORTED_AUDIO_EXTS,
        file_description_text="audio files",
        search_subdirectories=False,
        print_errors=False,
    )

    if len(all_subs) == 0:
        CliUtils.print_warning("No supported subtitles in directory. Supported formats:")
        CliUtils.print_rich(SUPPORTED_SUB_EXTS)
        return result

    if (len(all_videos) + len(all_audio)) == 0:
        CliUtils.print_warning("No media files in directory. Suppported formats:")
        CliUtils.print_rich(SUPPORTED_VIDEO_EXTS + SUPPORTED_AUDIO_EXTS)
        return result

    # Try to match a video file first. If we can't find it, move on to looking at audio files
    for subs in all_subs:
        sub_stem = subs.stem.split(".")[0]

        matched = False
        for video in all_videos:
            video_stem = video.stem.split(".")[0]

            if sub_stem == video_stem:
                matched = True
                result[subs] = video
                break

        if matched:
            continue

        for audio in all_audio:
            audio_stem = audio.stem.split(".")[0]

            if sub_stem == audio_stem:
                matched = True
                result[subs] = audio
                break

    if len(result) == 0:
        CliUtils.print_warning("Could not match subtitles to media files.")

    return result


def extract_frame(
    video_path: Path,
    output_directory: Path,
    timestamp_ms: int,
    extract_log: list[str] = [],
    file_format: str = "jpg",
) -> Path:
    """Extract a screenshot from a specific position in a video"""
    # Create output filename
    output_path = generate_output_path(video_path, output_directory, str(timestamp_ms), file_format)

    # For efficiency, don't extract twice if we have already extracted it
    if str(output_path) in extract_log:
        return output_path

    # Convert timestamp from milliseconds to seconds
    timestamp_seconds = timestamp_ms / 1000.0

    # Extract
    (
        ffmpeg.input(str(video_path), ss=timestamp_seconds)
        .output(str(output_path), vframes=1, loglevel="quiet")
        .run(overwrite_output=True)
    )

    return output_path


def extract_audio_segment(
    video_path: Path,
    output_directory: Path,
    start_time_ms: int,
    end_time_ms: int,
    extract_log: list[str] = [],
    file_format: str = "mp3",
):
    """Extract an audio clip from a media file and normalise it for speech."""

    # Create output filename
    output_path = generate_output_path(
        video_path, output_directory, str(start_time_ms), file_format
    )

    # For efficiency, don't extract twice if we have already extracted it
    if str(output_path) in extract_log:
        return output_path

    # Convert start and end times from milliseconds to seconds
    start_time_seconds = start_time_ms / 1000.0
    end_time_seconds = end_time_ms / 1000.0

    # Extract it
    (
        ffmpeg.input(str(video_path.resolve()), ss=start_time_seconds, to=end_time_seconds)
        .filter("speechnorm")
        .output(str(output_path.resolve()), format=file_format, loglevel="quiet")
        .run(overwrite_output=True)
    )

    return output_path



import platform
import shlex
import subprocess
from shutil import which

from .cli_utils import CliUtils


def get_platform() -> str:
    return platform.system()


def program_exists(tool_name: str) -> bool:
    """Check if a program can be accessed from the command line"""
    result = which(tool_name) is not None
    return result


# Mothballed for now, causes the package to become corrupted
# Fixed using installer for now
#
# def install_package(
#     package_name: str | list[str], app_name: str, package_index: str | None = None
# ) -> None:
#     # Convert to string if we have a list
#     if isinstance(package_name, list):
#         package_name = " ".join(package_name)

#     update_command: str | None = None
#     if program_exists("uv"):
#         result = get_command_output("uv tool list")
#         # Double check that we have installed it with uv
#         if app_name in result:
#             update_command = f"uv tool update {app_name} --with {package_name}"

#             if package_index:
#                 update_command += f" --index {package_index}"

#     if update_command:
#         run_command(update_command, True)
#     else:
#         CliUtils.print_warning(f"Not installing {package_name}, manual installation detected.")


def update_package(package_name: str, app_name: str) -> None:
    """Update a python package within the uv tool package"""

    update_command: str | None = None
    if program_exists("uv"):
        result = get_command_output("uv tool list")
        # Double check that we have installed it with uv
        if app_name in result:
            update_command = f"uv tool upgrade {app_name} --upgrade-package {package_name}"

    if update_command:
        run_command(update_command, True)
    else:
        CliUtils.print_warning(f"Not updating {package_name}, manual installation detected.")


def run_command(command: str | list[str], print_command: bool = False) -> None:
    """Run a command line program"""

    if isinstance(command, str):
        command = arg_string_to_list(command)

    if print_command:
        command_string = " ".join(command)
        CliUtils.print_status(f"Running command: {command_string}")

    if get_platform() == "Windows":
        process = subprocess.Popen(command, shell=True)
    else:
        process = subprocess.Popen(command)

    process.wait()


def arg_string_to_list(argument_string: str) -> list[str]:
    """Converts a string of arguments into a list"""

    arg_list = shlex.split(argument_string)
    return arg_list


def get_command_output(command: str | list[str]) -> str:
    if isinstance(command, str):
        command = arg_string_to_list(command)

    try:
        if get_platform() == "Windows":
            runner = subprocess.run(command, capture_output=True, text=True, shell=True)
        else:
            runner = subprocess.run(command, capture_output=True, text=True)

        result = runner.stdout

        return str(result)
    except Exception:
        return ""



import random
from pathlib import Path

import genanki
import pandas as pd

from .cli_utils import CliUtils
from .config import get_resources_directory
from .frequency_analyser import frequency_analyser
from .lookup import copy_all_word_audio, lookup_all_words
from .translator import translate_all_sentences
from .utils import (
    dataframe_to_excel,
    generate_output_path,
    load_json,
    read_file_to_string,
    sanitise_string_html,
)
from .video_manipulation import extract_sentence_media, match_subtitles_to_media


def create_anki(
    input_directory: Path,
    language: str,
    subs_offset: int,
    subs_buffer: int,
    max_cards: int,
    lemma: bool,
    filter_stop_words: bool,
    extract_media: bool,
    include_no_definition: bool,
    native_language: str,
    word_audio_path: Path | None = None,
    dictionary_path: Path | None = None,
    excluded_words: Path | None = None,
) -> Path | None:
    """Main entry point for the anki deck builder"""

    # Config
    CliUtils.print_status("Anki builder: configuring paths")
    anki_media_dir = input_directory / "media"
    anki_sentence_audio_dir = anki_media_dir / "sentence_audio"
    anki_screenshot_dir = anki_media_dir / "screenshots"
    anki_word_dir = anki_media_dir / "word_audio"
    anki_media_dir.mkdir(parents=True, exist_ok=True)
    anki_sentence_audio_dir.mkdir(parents=True, exist_ok=True)
    anki_screenshot_dir.mkdir(parents=True, exist_ok=True)
    anki_word_dir.mkdir(parents=True, exist_ok=True)

    deck_output_path = generate_output_path(input_directory, anki_media_dir, "deck", "apkg")
    deck_name = deck_output_path.stem

    # Frequency analysis
    CliUtils.print_status("Anki builder: running frequency analysis")
    df = frequency_analyser(
        input_path=input_directory,
        output_directory=anki_media_dir,
        language=language,
        lemmatise_words=lemma,
        filter_stop_words=filter_stop_words,
        exclude_spreadsheet_path=excluded_words,
    )

    if df is None:
        return None

    CliUtils.print_status("Anki builder: preparing dataframe")
    # Add in additional columns that we will use to store information for the Anki cards
    df = df.assign(screenshot_path="", sentence_audio_path="", word_audio_path="")
    df = df.assign(definition="", translation="")

    # Get the definitions. Do this before everything else since we want to limit cards asap
    if dictionary_path:
        CliUtils.print_status("Anki builder: getting definitions")
        df = lookup_all_words(df=df, dictionary_path=dictionary_path)

    CliUtils.print_status("Anki builder: pruning cards")
    # If specified by user, remove cards without a definition
    if not include_no_definition:
        df = filter_no_definition(df=df)

    # Limit the number of cards
    df = limit_cards(df=df, max_cards=max_cards)

    # Work out the media files for each subtitle
    CliUtils.print_status("Anki builder: matching subtitles to media files")
    subs_to_media = match_subtitles_to_media(input_directory)
    if len(subs_to_media) == 0:
        return None

    # Get sentence audio and screenshots and update dataframe with paths
    if extract_media:
        CliUtils.print_status("Anki builder: extracting media to use in cards")
        df = extract_sentence_media(
            audio_dir=anki_sentence_audio_dir,
            screenshot_dir=anki_screenshot_dir,
            subs_to_media=subs_to_media,
            df=df,
            subs_offset=subs_offset,
            subs_buffer=subs_buffer,
        )

    # Get the word audio
    if word_audio_path:
        CliUtils.print_status("Anki builder: getting example audio for individual words")
        df = copy_all_word_audio(
            df=df, library_path=word_audio_path, output_directory=anki_word_dir
        )

    CliUtils.print_status("Getting translations")
    df = translate_all_sentences(df=df, from_lang=language, to_lang=native_language)

    # Build deck
    CliUtils.print_status("Building deck")
    anki_notes, media_files = produce_anki_notes(
        df=df, deck_name=deck_name, include_no_definition=include_no_definition
    )

    package = create_package(deck_name=deck_name, anki_notes=anki_notes, media_files=media_files)

    # Save the df and the Anki package
    CliUtils.print_status("Anki builder: saving data")
    write_package(deck_output_path, package)
    save_deck_definition_sheet(df, input_directory, anki_media_dir)

    return None


def limit_cards(df: pd.DataFrame, max_cards: int) -> pd.DataFrame:
    """Limit cards in a deck to a specified value by modifying the dataframe used to build it"""
    df = df.head(max_cards)
    return df


def filter_no_definition(df: pd.DataFrame) -> pd.DataFrame:
    """Remove cards that have a blank entry for definiton"""
    df = df[df["definition"] != ""]

    return df


def write_package(file_path: Path, package: genanki.Package) -> None:
    """Save Anki package to file"""
    package.write_to_file(file_path)


def create_package(deck_name, anki_notes: list, media_files: list) -> genanki.Package:
    """Build the Anki package in memory"""

    deck = create_deck(
        deck_id=generate_anki_id(),
        deck_name=deck_name,
        note_cards=anki_notes,
    )

    package = genanki.Package(deck)

    if len(media_files) > 0:
        package.media_files = media_files

    return package


def produce_anki_notes(
    df: pd.DataFrame, deck_name: str, include_no_definition: bool
) -> tuple[list, list]:
    """Convert the data from the dataframe into Anki notes. Return the required media files as the second variable in the tuple"""

    # Configure paths
    html_path = get_resources_directory() / "html" / "anki"
    css_path = html_path / "anki_card.css"
    front_path = html_path / "front.html"
    back_path = html_path / "back.html"
    model_fields_path = html_path / "fields.json"

    # Build model
    model_fields = load_json(model_fields_path)
    assert isinstance(model_fields, list)

    model_template = [
        {
            "name": deck_name,
            "qfmt": read_file_to_string(front_path),
            "afmt": read_file_to_string(back_path),
        },
    ]

    model = create_model(
        model_id=generate_anki_id(),
        model_name=deck_name,
        fields=model_fields,
        template=model_template,
        css=read_file_to_string(css_path),
    )

    i = 0
    note_cards = []
    media_files = []
    for _, row in df.iterrows():
        # Get data from row
        word = str(row["word"])
        example = str(row["example"])
        word_frequency = str(row["frequency"])
        image_path = Path(str(row["screenshot_path"]))
        sentence_audio_path = Path(str(row["sentence_audio_path"]))
        word_audio_path = Path(str(row["word_audio_path"]))
        definition = str(row["definition"])
        translation = str(row["translation"])
        card_index = str(i)

        # Put the file names for the audio into Anki card format
        sentence_audio_field_anki = format_audio_field(sentence_audio_path.name)
        word_audio_field_anki = format_audio_field(word_audio_path.name)
        image_field_anki = format_image_field(image_path.name)

        # Set the field values for the card
        note_fields = [
            card_index,
            sanitise_string_html(word),
            sanitise_string_html(example),
            sanitise_string_html(translation),
            definition,
            word_frequency,
            sentence_audio_field_anki,
            word_audio_field_anki,
            image_field_anki,
        ]

        # Create an Anki note
        note = create_note(model, note_fields)
        note_cards.append(note)

        # Media files
        if sentence_audio_path.exists() and sentence_audio_path.is_file():
            media_files.append(sentence_audio_path)

        if word_audio_path.exists() and word_audio_path.is_file():
            media_files.append(word_audio_path)

        if image_path.exists() and image_path.is_file():
            media_files.append(image_path)

        i += 1

    return note_cards, media_files


def save_deck_definition_sheet(df: pd.DataFrame, input_path: Path, output_path: Path):
    """Save the dataframe used to build the Anki deck to file"""

    save_path = generate_output_path(input_path, output_path, "deck_resources", "xlsx")
    dataframe_to_excel(df, save_path)


def create_note(model: genanki.Model, fields: list) -> genanki.Note:
    """Generate an individual Anki note"""
    note = genanki.Note(model=model, fields=fields)
    return note


def create_model(
    model_id: int, model_name: str, fields: list, template: list, css: str = ""
) -> genanki.Model:
    """Create an Anki model with the fields and card formatting"""
    model = genanki.Model(
        model_id=model_id, name=model_name, fields=fields, templates=template, css=css
    )

    return model


def create_deck(deck_id: int, deck_name: str, note_cards: list) -> genanki.Deck:
    """In memory, add Anki notes to a deck object"""
    deck = genanki.Deck(deck_id=deck_id, name=deck_name)

    if len(note_cards) > 0:
        for note in note_cards:
            deck.add_note(note)

    return deck


def generate_anki_id() -> int:
    """Generate a random number to act as the Anki deck ID"""
    model_id = random.randrange(1 << 30, 1 << 31)

    return model_id


def format_audio_field(file_name: str) -> str:
    """Format the file name so that it can be used in an Anki field. Input string should only be the file stem plus file file suffix.
    Note to self. You cannot specify audio/sound.mp3 as Anki has a flat file structure.
    All media files must therefore have unique names
    """
    if file_name == "":
        return ""

    output_string = f"[sound:{file_name}]"

    return output_string


def format_image_field(file_name: str) -> str:
    """Format the file name so that it can be used in an Anki field. Input string should only be the file stem plus file file suffix.
    Note to self. You cannot specify images/image.jpg as Anki has a flat file structure.
    All media files must therefore have unique names
    """
    if file_name == "":
        return ""

    output_string = f"<img src='{file_name}'>"

    return output_string



import re
from pathlib import Path
from typing import Any

import typer
from rich import print
from rich.text import Text


class CliUtils:
    """Utilities to make life easier with a typer cli application"""

    @staticmethod
    def clean_string(input_str: str) -> str:
        """Callback that strips whitespace from string"""
        cleaned = input_str.strip()

        return cleaned

    @staticmethod
    def print_rich(item: Any) -> None:
        """Print an object using rich's formatting"""
        print(item)

    @staticmethod
    def print_status(message: str) -> None:
        """Print status message using rich"""
        print(
            Text(
                f"~~~ {message} ~~~",
                style="bold blue",
            )
        )

    @staticmethod
    def print_warning(message: str) -> None:
        """Print warning message using rich"""
        print(
            Text(
                f"### {message} ###",
                style="bold red",
            )
        )

    @staticmethod
    def print_error(message: str, e: Exception | None = None) -> None:
        """Print error message using rich. Optionally print exception text"""

        error_text = ""
        if e is not None:
            error_text = f"{str(e)}"

        print(
            Text(
                f"### {message} ###\n{error_text}",
                style="bold red",
            )
        )

    @staticmethod
    def validate_language_code(lang: str) -> str:
        """Forces language code to lower case without trailing whitespace,"""
        lang = lang.strip().lower()

        if len(lang) != 2:
            raise typer.BadParameter(
                f"Language must be a two letter code. Run `gogoprimer --list-langages` for a list of supported languages. User provided language: `{lang}`."
            )

        return lang

    @staticmethod
    def validate_optional_directory(path: Path | None) -> Path | None:
        """Checks that a directory path is theoretically valid, if specified"""
        if path is None:
            return None
        else:
            return CliUtils.validate_directory(path)

    @staticmethod
    def validate_optional_file(path: Path | None) -> Path | None:
        """Checks that a file path is theoretically valid, if specified"""
        if path is None:
            return None
        else:
            return CliUtils.validate_file(path)

    @staticmethod
    def validate_directory(path: Path) -> Path:
        """Checks that a directory path is theoretically valid"""
        return CliUtils._validate_path(path, must_be_directory=True)

    @staticmethod
    def validate_file(path: Path) -> Path:
        """Checks that a file path is theoretically valid"""
        return CliUtils._validate_path(path, must_be_file=True)

    @staticmethod
    def validate_path_exists(path: Path) -> Path:
        """Checks that a path exists on the user's system"""
        return CliUtils._validate_path(path, must_exist=True)

    @staticmethod
    def _validate_path(
        path: Path,
        allow_blank: bool = False,
        must_be_file: bool = False,
        must_be_directory: bool = False,
        must_exist: bool = False,
    ) -> Path:
        """Helper method to enable various checks"""
        if (not allow_blank) and (path is None):
            raise typer.BadParameter("Cannot be blank!")
        if (must_be_file) and (not path.is_file()):
            raise typer.BadParameter(f"Must be a file path. User provided path: `{path}`.")
        if (must_be_directory) and (not path.is_dir()):
            raise typer.BadParameter(
                f"Must be a valid directory (aka folder). User provided path: `{path}`."
            )
        if (must_exist) and (not path.exists()):
            raise typer.BadParameter(f"Path does not exist. User provided path: `{path}`.")

        return path

    @staticmethod
    def validate_url(url: str) -> str:
        """Validate that a url is theoretically possible"""
        regex = re.compile(
            r"(\w+://)?"  # protocol                      (optional)
            r"(\w+\.)?"  # host                          (optional)
            r"(([\w-]+)\.(\w+))"  # domain
            r"(\.\w+)*"  # top-level domain              (optional, can have > 1)
            r"([\w\-\._\~/]*)*(?<!\.)"  # path, params, anchors, etc.   (optional)
        )

        url = url.strip()

        if url == "":
            raise typer.BadParameter("URL cannot be blank!")
        if not regex.match(url):
            raise typer.BadParameter(f"Invalid url. User provided url: `{url}`.")

        return url



from lemon_tizer import LemonTizer

from .help_text import SupportedLanguages


def lemma_dummy(language: str) -> None:
    """Function to initialise the lemmatiser on first run and download the required modules"""
    lt = LemonTizer(language, "lg")
    lt.lemmatize_sentence("Wololo")


def language_supported(language: str) -> bool:
    """Check if the language is supported.
    LemonTizer has its own function that is more accurate but this works offline."""

    supported = language in SupportedLanguages.spacy_languages.keys()
    return supported



import importlib.metadata
import re
from pathlib import Path
from typing import Any

import rtoml
import typer

from .cli_utils import CliUtils
from .command_runner import get_platform, program_exists, run_command
from .help_text import HelpText

"""Metadata"""
APP_NAME = "gogadget"

"""Supported file formats"""
SUPPORTED_VIDEO_EXTS = [
    ".mp4",
    ".webm",
    ".mov",
    ".mkv",
    ".mpeg",
    ".mpg",
    ".avi",
    ".ogv",
    ".wmv",
    ".m4v",
    ".3gp",
    ".ts",
]
SUPPORTED_AUDIO_EXTS = [".mp3", ".ogg", ".wav", ".opus", ".m4a", ".aac", ".aiff", ".flac"]
SUPPORTED_SUB_EXTS = [".srt", ".vtt"]
SUPPORTED_WORD_AUDIO_EXTS = [".mp3"]


def get_version_number() -> str:
    """Gets the version number of the python package"""
    try:
        version = importlib.metadata.version(f"{APP_NAME}")
        return f"{APP_NAME} version: {version}"
    except Exception:
        return "[bold red]Couldn't get version. If you are running this from source, no version number is available."


def main_package_directory() -> Path:
    """Get the path of where the python source files are for the project"""
    # Just return the directory of config.py since it's in the root of the package
    path = Path(__file__).parent.resolve()
    return path


def get_resources_directory() -> Path:
    """Get the path of the project (source) resources file"""
    path = main_package_directory() / "resources"

    return path


class ConfigFile:
    """Class to configure the user config file"""

    # User configuration directory e.g. ~/Library/Application Support/... on macos
    CONFIG_DIRECTORY = Path(typer.get_app_dir(APP_NAME))

    def config_file_exists(self) -> bool:
        """Has the config file been created?"""
        config_file = self.get_config_file_path()
        return config_file.exists()

    def launch_config_file_in_editor(self) -> None:
        """Open in the platform default editor (macos/linux) On Windows, try to open in vscode, fall back to notepad"""
        config_file = self.get_config_file_path()

        if not config_file.exists():
            self.create_config_file()

        config_file_str = str(config_file.resolve())

        if get_platform() == "Windows":
            # Try to open with vscode. If not available, open in notepad
            if program_exists("code"):
                run_command(["code", config_file_str])
            else:
                run_command(["notepad.exe", config_file_str])
        else:
            typer.launch(config_file_str)

    def get_config_file_path(self) -> Path:
        """Get the default path name for the config file"""
        config_file = ConfigFile.CONFIG_DIRECTORY / f"{APP_NAME}.toml"

        return config_file

    def create_config_file(self) -> Path:
        """Create a blank config file if it doesn't already exist"""
        config_file = self.get_config_file_path()

        if config_file.exists():
            return config_file

        config_root = config_file.parent

        if not config_root.exists():
            config_root.mkdir(parents=True, exist_ok=True)

        config_file.touch()

        return config_file

    def factory_reset(self) -> None:
        """Factory reset of app defaults.

        Deletes the config file to avoid accidentally writing old data. It will be regenerated next time the app is launched
        """

        config_file = self.get_config_file_path()

        if config_file.exists() and config_file.is_file():
            config_file.unlink()
            CliUtils.print_rich(
                "Config file deleted. Will be re-generated next time the program is launched."
            )

    def write_defaults(self) -> None:
        """Write the factory defaults to the config file in .toml format"""

        # Since this is run on every load but rarely used, only import within the function
        import tomlkit
        from tomlkit.items import Table

        # Helper function to pull together a table (section) of the toml file
        # Write everything as a string to make it easier for novices to edit. We will convert it later.
        def create_table(content: dict[str, Any] | None, comment_str: str | None = None) -> Table:
            result = tomlkit.table()

            if comment_str:
                result.add(tomlkit.comment(comment_str))
            if content:
                for key, value in content.items():
                    if value is None:
                        value = ""

                    # Escape any paths
                    if isinstance(value, Path):
                        value = re.sub(r"\\", r"\\\\", str(value))

                    result.add(key, str(value))

            return result

        # Create config file if it doesn't currently exist
        file_path = self.create_config_file()

        doc = tomlkit.document()
        doc.add("instructions", create_table(None, HelpText.toml_instructions))
        doc.add(
            "general", create_table(self.get_object_values(self.general), HelpText.toml_general)
        )
        doc.add(
            "external_resources",
            create_table(self.get_object_values(self.external_resources), HelpText.toml_external),
        )
        doc.add("anki", create_table(self.get_object_values(self.anki), HelpText.toml_anki))
        doc.add(
            "lemmatiser",
            create_table(self.get_object_values(self.lemmatiser), HelpText.toml_lemmatiser),
        )
        doc.add(
            "downloader",
            create_table(self.get_object_values(self.downloader), HelpText.toml_downloader),
        )
        doc.add(
            "transcriber",
            create_table(self.get_object_values(self.transcriber), HelpText.toml_transcriber),
        )

        with file_path.open("w", encoding="utf-8") as f:
            tomlkit.dump(doc, f)

    def get_object_values(self, defaults_class: object) -> dict[str, Any]:
        """Get the non-built in parameters from an object

        Be careful if the object has both parameters and functions.
        This is designed to work with objects with parameters only.
        """

        output: dict[str, Any] = {}
        for key, value in vars(defaults_class).items():
            if not key.startswith("__"):
                if isinstance(value, Path):
                    value = str(value)
                output[key] = value

        return output

    def read_defaults(self) -> None:
        """Read the defaults values from a toml file"""
        config_file = self.get_config_file_path()

        if not config_file.exists():
            CliUtils.print_error("Could not read configuration file. Defaults have not been loaded")
            return None

        config: dict = {}

        parse_error = False
        try:
            config = rtoml.load(config_file)
        except rtoml.TomlParsingError:
            parse_error = True

        # Attempt to fix it, most likely culprit is escaped paths
        if parse_error:
            try:
                self.fix_toml_paths(config_file)
                config = rtoml.load(config_file)
            except Exception as e:
                CliUtils.print_error(
                    "Could not read user config file. gogadget will use its own default values.\nYou may want to run `gogadget set-defaults --factory` to reset the user configuration.",
                    e,
                )
                return None

        # We don't want this to cause a crash as we need the user to always be able to run gogadget set-defaults --factory
        # This runs every time, including before that command
        # If there is an issue, the tool will revert to the defaults set in the classes
        try:
            general = config["general"]
            self.general.language = self.read_str(general, "language", self.general.language)
            self.general.language_for_translations = self.read_str(
                general, "language_for_translations", self.general.language_for_translations
            )
            self.general.output_directory = self.read_path(
                general, "output_directory", self.general.output_directory, blank_is_none=False
            )  # type: ignore

            external_resources = config["external_resources"]
            self.external_resources.dictionary_file = self.read_path(
                external_resources, "dictionary_file", self.external_resources.dictionary_file
            )
            self.external_resources.word_audio_directory = self.read_path(
                external_resources,
                "word_audio_directory",
                self.external_resources.word_audio_directory,
            )
            self.external_resources.word_exclude_spreadsheet = self.read_path(
                external_resources,
                "word_exclude_spreadsheet",
                self.external_resources.word_exclude_spreadsheet,
            )

            anki = config["anki"]
            self.anki.extract_media = self.read_bool(anki, "extract_media", self.anki.extract_media)
            self.anki.include_words_with_no_definition = self.read_bool(
                anki, "include_words_with_no_definition", self.anki.include_words_with_no_definition
            )
            self.anki.subs_offset_ms = self.read_int(
                anki, "subs_offset_ms", self.anki.subs_offset_ms
            )
            self.anki.subs_buffer_ms = self.read_int(
                anki, "subs_buffer_ms", self.anki.subs_buffer_ms
            )
            self.anki.max_cards_in_deck = self.read_int(
                anki, "max_cards_in_deck", self.anki.max_cards_in_deck
            )

            lemmatiser = config["lemmatiser"]
            self.lemmatiser.lemmatise = self.read_bool(
                lemmatiser, "lemmatise", self.lemmatiser.lemmatise
            )
            self.lemmatiser.filter_out_non_alpha = self.read_bool(
                lemmatiser, "filter_out_non_alpha", self.lemmatiser.filter_out_non_alpha
            )
            self.lemmatiser.filter_out_stop_words = self.read_bool(
                lemmatiser, "filter_out_stop_words", self.lemmatiser.filter_out_stop_words
            )
            self.lemmatiser.convert_input_to_lower = self.read_bool(
                lemmatiser, "convert_input_to_lower", self.lemmatiser.convert_input_to_lower
            )
            self.lemmatiser.convert_output_to_lower = self.read_bool(
                lemmatiser, "convert_output_to_lower", self.lemmatiser.convert_output_to_lower
            )
            self.lemmatiser.return_just_first_word_of_lemma = self.read_bool(
                lemmatiser,
                "return_just_first_word_of_lemma",
                self.lemmatiser.return_just_first_word_of_lemma,
            )

            downloader = config["downloader"]
            self.downloader.advanced_options = self.read_str(
                downloader, "advanced_options", self.downloader.advanced_options
            )
            self.downloader.format = self.read_str(downloader, "format", self.downloader.format)
            self.downloader.subtitle_language = self.read_str(
                downloader, "subtitle_language", self.downloader.subtitle_language
            )

            transcriber = config["transcriber"]
            self.transcriber.whisper_model = self.read_str(
                transcriber, "whisper_model", self.transcriber.whisper_model
            )
            self.transcriber.alignment_model = self.read_str(
                transcriber, "alignment_model", self.transcriber.alignment_model
            )
            self.transcriber.subtitle_format = self.read_str(
                transcriber, "subtitle_format", self.transcriber.subtitle_format
            )
            self.transcriber.whisper_use_gpu = self.read_bool(
                transcriber, "whisper_use_gpu", self.transcriber.whisper_use_gpu
            )

        except Exception as e:
            CliUtils.print_error(
                "Could not read configuration parameters.\nPlease fix your user config file and / or consider running `gogadget set-defaults --factory` to reset it.",
                e,
            )

    def fix_toml_paths(self, file_path: Path) -> None:
        """Escapes any windows paths in the toml file"""
        # Read the content of the file
        with file_path.open("r", encoding="utf-8") as f:
            content = f.read()

        # Replace single backslashes with double backslashes, ignoring existing double backslashes
        escaped_content = re.sub(r"(?<!\\)\\(?!\\)", r"\\\\", content)

        # Write the modified content back to the file
        with file_path.open("w", encoding="utf-8") as f:
            f.write(escaped_content)

    def read_path(
        self, toml_dict: dict, dict_key: str, default_value: Path | None, blank_is_none: bool = True
    ) -> Path | None:
        value = toml_dict.get(dict_key)
        """Read value from a dictionary and convert it to path. If it's blank, change return type based upon blank_is_none"""

        if value is None:
            # Nothing found in the dictionary, return the default value, regardless of value of blank_is_none
            return default_value

        # Convert whatever we have to a string and remove all whitespace
        value = str(value).strip()

        if value != "":
            # Not blank, convert it to a path and return
            return Path(value)
        elif blank_is_none:
            # It's blank and we have decided that blank should be treated as none
            return None
        else:
            # Otherwise, blank is blank
            # Path("") is actually equivalent to Path(".")
            # Therefore, let's be safe and just return the default value
            return default_value

    def read_bool(self, toml_dict: dict, dict_key: str, default_value: bool) -> bool:
        """Get a value from a dictionary. If it's a bool, return it. If not, try to convert the string into the name of a bool type"""
        value = toml_dict.get(dict_key, default_value)

        if isinstance(value, bool):
            return value

        if isinstance(value, str):
            value = value.lower().strip()

            if value == "true":
                return True
            elif value == "false":
                return False

        # Unclear what we have, just return the default
        return default_value

    def read_int(self, toml_dict: dict, dict_key: str, default_value: int) -> int:
        """Get a value from a dictionary and convert to int. If key not present, return default. Return default if we can't convert it to int."""
        try:
            value = int(toml_dict.get(dict_key, default_value))
            return value
        except ValueError:
            return default_value

    def read_str(self, toml_dict: dict, dict_key: str, default_value: str) -> str:
        """Get a value from a dictionary and convert to str and strip whitespace. Return default value if key isn't present in dict."""
        value = str(toml_dict.get(dict_key, default_value)).strip()
        return value

    class general:
        """Object to hold default values. Pre-initialised with factory defaults"""

        language: str = ""
        language_for_translations: str = "en"
        output_directory: Path = Path("")

    class external_resources:
        """Object to hold default values. Pre-initialised with factory defaults"""

        word_exclude_spreadsheet: Path | None = None
        dictionary_file: Path | None = None
        word_audio_directory: Path | None = None

    class anki:
        """Object to hold default values. Pre-initialised with factory defaults"""

        extract_media: bool = True
        include_words_with_no_definition: bool = True
        subs_offset_ms: int = 0
        subs_buffer_ms: int = 50
        max_cards_in_deck: int = 100

    class lemmatiser:
        """Object to hold default values. Pre-initialised with factory defaults"""

        lemmatise: bool = True
        filter_out_non_alpha: bool = True
        filter_out_stop_words: bool = True
        convert_input_to_lower: bool = True
        convert_output_to_lower: bool = True
        return_just_first_word_of_lemma: bool = True

    class downloader:
        """Object to hold default values. Pre-initialised with factory defaults"""

        advanced_options: str = ""
        format: str = ""
        subtitle_language: str = ""

    class transcriber:
        """Object to hold default values. Pre-initialised with factory defaults"""

        whisper_model: str = "deepdml/faster-whisper-large-v3-turbo-ct2"
        alignment_model: str = ""
        subtitle_format: str = "vtt"
        whisper_use_gpu: bool = False



"""
Taken from: https://github.com/yt-dlp/yt-dlp/blob/master/devscripts/cli_to_api.py
Public domain licence: https://raw.githubusercontent.com/yt-dlp/yt-dlp/refs/heads/master/LICENSE

File not been modified from source, aside from this comment and removal of if __name__ == "__main__" block.

Used to convert yt-dlp command line arguments into a dictionary that can be read by the python api.
"""

# Allow direct execution
import os
import sys

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

import yt_dlp
import yt_dlp.options

create_parser = yt_dlp.options.create_parser


def parse_patched_options(opts):
    patched_parser = create_parser()
    patched_parser.defaults.update(
        {
            "ignoreerrors": False,
            "retries": 0,
            "fragment_retries": 0,
            "extract_flat": False,
            "concat_playlist": "never",
        }
    )
    yt_dlp.options.create_parser = lambda: patched_parser
    try:
        return yt_dlp.parse_options(opts)
    finally:
        yt_dlp.options.create_parser = create_parser


default_opts = parse_patched_options([]).ydl_opts


def cli_to_api(opts, cli_defaults=False):
    opts = (yt_dlp.parse_options if cli_defaults else parse_patched_options)(opts).ydl_opts

    diff = {k: v for k, v in opts.items() if default_opts[k] != v}
    if "postprocessors" in diff:
        diff["postprocessors"] = [
            pp for pp in diff["postprocessors"] if pp not in default_opts["postprocessors"]
        ]
    return diff



from importlib import import_module
from pathlib import Path

import typer
from typing_extensions import Annotated

from .cli_utils import CliUtils
from .config import (
    APP_NAME,
    ConfigFile,
    get_resources_directory,
    get_version_number,
)
from .help_text import HelpText, SupportedLanguages, ffmpeg_warning

"""
Define settings for the cli framework (Typer) and load defaults from config file
"""
app = typer.Typer(
    no_args_is_help=True,
    rich_markup_mode="rich",
    add_completion=False,
)

CONFIG = ConfigFile()

# Write the config if it doesn't already exist
if not CONFIG.config_file_exists():
    CONFIG.write_defaults()

CONFIG.read_defaults()


"""
Primary commands
"""


@app.command(
    no_args_is_help=True,
    help=HelpText.create_anki_deck,
    rich_help_panel="Primary Functions",
    epilog=ffmpeg_warning(),
)
def anki_deck(
    input_directory: Annotated[
        Path,
        typer.Option(
            "--input",
            "-i",
            callback=CliUtils.validate_directory,
            help=HelpText.input_directory_anki,
            show_default=False,
            rich_help_panel="Required",
        ),
    ],
    language: Annotated[
        str,
        typer.Option(
            "--language",
            "-l",
            help=HelpText.language_code,
            callback=CliUtils.validate_language_code,
            rich_help_panel="Required",
        ),
    ] = CONFIG.general.language,
    translation_language: Annotated[
        str,
        typer.Option(
            "--translation-language",
            "-t",
            help=HelpText.language_code_translation,
            callback=CliUtils.validate_language_code,
            rich_help_panel="Optional",
        ),
    ] = CONFIG.general.language_for_translations,
    offset: Annotated[
        int,
        typer.Option("--offset", "-f", help=HelpText.subtitle_offset, rich_help_panel="Optional"),
    ] = CONFIG.anki.subs_offset_ms,
    buffer: Annotated[
        int,
        typer.Option("--buffer", "-b", help=HelpText.subtitle_buffer, rich_help_panel="Optional"),
    ] = CONFIG.anki.subs_buffer_ms,
    max_cards: Annotated[
        int,
        typer.Option("--max-cards", "-x", help=HelpText.max_cards, rich_help_panel="Optional"),
    ] = CONFIG.anki.max_cards_in_deck,
    word_audio: Annotated[
        Path | None,
        typer.Option(
            "--word-audio",
            "-w",
            callback=CliUtils.validate_optional_directory,
            help=HelpText.word_examples,
            rich_help_panel="Optional",
        ),
    ] = CONFIG.external_resources.word_audio_directory,
    dictionary: Annotated[
        Path | None,
        typer.Option(
            "--dictionary",
            "-d",
            callback=CliUtils.validate_optional_file,
            help=HelpText.dictionary_file,
            rich_help_panel="Optional",
        ),
    ] = CONFIG.external_resources.dictionary_file,
    excluded_words: Annotated[
        Path | None,
        typer.Option(
            "--excluded-words",
            "-e",
            callback=CliUtils.validate_optional_file,
            help=HelpText.exclude_sheet,
            rich_help_panel="Optional",
        ),
    ] = CONFIG.external_resources.word_exclude_spreadsheet,
    lemma: Annotated[
        bool,
        typer.Option(
            "--lemma/--no-lemma", "-m/-n", help=HelpText.lemmatise, rich_help_panel="Optional Flags"
        ),
    ] = CONFIG.lemmatiser.lemmatise,
    stop_words: Annotated[
        bool,
        typer.Option(
            "--stop-words/--no-stop-words",
            "-s/-p",
            help=HelpText.stop_words,
            rich_help_panel="Optional Flags",
        ),
    ] = (not CONFIG.lemmatiser.filter_out_stop_words),
    media: Annotated[
        bool,
        typer.Option(
            "--media/--no-media",
            "-q/-r",
            help=HelpText.anki_media,
            rich_help_panel="Optional Flags",
        ),
    ] = CONFIG.anki.extract_media,
    include_no_definition: Annotated[
        bool,
        typer.Option(
            "--include-no-definition/--exclude-no-definition",
            "-g/-h",
            help=HelpText.include_no_definition,
            rich_help_panel="Optional Flags",
        ),
    ] = CONFIG.anki.include_words_with_no_definition,
):
    CliUtils.print_status("Anki builder: starting")

    # Build deck
    anki = import_module(".create_anki", APP_NAME)
    anki.create_anki(
        input_directory=input_directory,
        language=language,
        native_language=translation_language,
        subs_offset=offset,
        subs_buffer=buffer,
        max_cards=max_cards,
        lemma=lemma,
        filter_stop_words=not (stop_words),
        include_no_definition=include_no_definition,
        extract_media=media,
        word_audio_path=word_audio,
        dictionary_path=dictionary,
        excluded_words=excluded_words,
    )


@app.command(
    no_args_is_help=True,
    rich_help_panel="Primary Functions",
    help=HelpText.download,
    epilog=ffmpeg_warning(),
)
def download(
    url: Annotated[
        str,
        typer.Option(
            "--url",
            "-i",
            help=HelpText.video_url,
            show_default=False,
            callback=CliUtils.validate_url,
            rich_help_panel="Required",
        ),
    ],
    output_directory: Annotated[
        Path,
        typer.Option(
            "--output",
            "-o",
            callback=CliUtils.validate_directory,
            help=HelpText.output_directory,
            rich_help_panel="Optional",
        ),
    ] = CONFIG.general.output_directory,
    format: Annotated[
        str,
        typer.Option(
            "--format",
            "-f",
            callback=CliUtils.clean_string,
            help=HelpText.video_format,
            rich_help_panel="Optional",
        ),
    ] = CONFIG.downloader.format,
    subtitle_language: Annotated[
        str,
        typer.Option(
            "--subtitle-language",
            "-l",
            callback=CliUtils.clean_string,
            help=HelpText.ytdlp_subtitles,
            rich_help_panel="Optional",
        ),
    ] = CONFIG.downloader.subtitle_language,
    advanced_options: Annotated[
        str,
        typer.Option(
            "--advanced-options",
            "-a",
            callback=CliUtils.clean_string,
            help=HelpText.ytdlp_options,
            rich_help_panel="Optional",
        ),
    ] = CONFIG.downloader.advanced_options,
):
    CliUtils.print_status("Downloader: starting")
    downloader = import_module(".downloader", APP_NAME)

    if subtitle_language:
        subtitle_language = subtitle_language.lower().strip()

    downloader.downloader(
        url=url,
        download_directory=output_directory,
        format=format,
        subtitle_language=subtitle_language,
        other_options=advanced_options,
    )


@app.command(
    no_args_is_help=True,
    rich_help_panel="Primary Functions",
    help=HelpText.download_audio,
    epilog=ffmpeg_warning(),
)
def download_audio(
    url: Annotated[
        str,
        typer.Option(
            "--url",
            "-i",
            help=HelpText.video_url,
            show_default=False,
            callback=CliUtils.validate_url,
            rich_help_panel="Required",
        ),
    ],
    output_directory: Annotated[
        Path,
        typer.Option(
            "--output",
            "-o",
            callback=CliUtils.validate_directory,
            help=HelpText.output_directory,
            rich_help_panel="Optional",
        ),
    ] = CONFIG.general.output_directory,
    advanced_options: Annotated[
        str,
        typer.Option(
            "--advanced-options",
            "-a",
            callback=CliUtils.clean_string,
            help=HelpText.ytdlp_options,
            rich_help_panel="Optional",
        ),
    ] = CONFIG.downloader.advanced_options,
):
    CliUtils.print_status("Downloader: starting")
    downloader = import_module(".downloader", APP_NAME)
    downloader.downloader(
        url=url,
        download_directory=output_directory,
        extract_audio=True,
        other_options=advanced_options,
    )


@app.command(
    no_args_is_help=True,
    rich_help_panel="Primary Functions",
    help=HelpText.download_subtitles,
    epilog=ffmpeg_warning(),
)
def download_subtitles(
    url: Annotated[
        str,
        typer.Option(
            "--url",
            "-i",
            help=HelpText.video_url,
            show_default=False,
            callback=CliUtils.validate_url,
            rich_help_panel="Required",
        ),
    ],
    subtitle_language: Annotated[
        str,
        typer.Option(
            "--subtitle-language",
            "-l",
            callback=CliUtils.clean_string,
            help=HelpText.ytdlp_subtitles_required,
            rich_help_panel="Required",
        ),
    ] = CONFIG.downloader.subtitle_language,
    output_directory: Annotated[
        Path,
        typer.Option(
            "--output",
            "-o",
            callback=CliUtils.validate_directory,
            help=HelpText.output_directory,
            rich_help_panel="Optional",
        ),
    ] = CONFIG.general.output_directory,
    advanced_options: Annotated[
        str,
        typer.Option(
            "--advanced-options",
            "-a",
            callback=CliUtils.clean_string,
            help=HelpText.ytdlp_options,
            rich_help_panel="Optional",
        ),
    ] = CONFIG.downloader.advanced_options,
):
    CliUtils.print_status("Downloader: starting")

    downloader = import_module(".downloader", APP_NAME)

    if subtitle_language:
        subtitle_language = subtitle_language.lower().strip()

    downloader.downloader(
        url=url,
        download_directory=output_directory,
        subs_only=True,
        other_options=advanced_options,
        subtitle_language=subtitle_language,
    )

    return


@app.command(
    no_args_is_help=True, help=HelpText.frequency_analysis, rich_help_panel="Primary Functions"
)
def frequency_analysis(
    input_path: Annotated[
        Path,
        typer.Option(
            "--input",
            "-i",
            help=HelpText.transcribe_path,
            show_default=False,
            callback=CliUtils.validate_path_exists,
            rich_help_panel="Required",
        ),
    ],
    language: Annotated[
        str,
        typer.Option(
            "--language",
            "-l",
            help=HelpText.language_code,
            callback=CliUtils.validate_language_code,
            rich_help_panel="Required",
        ),
    ] = CONFIG.general.language,
    output_directory: Annotated[
        Path,
        typer.Option(
            "--output",
            "-o",
            callback=CliUtils.validate_directory,
            help=HelpText.output_directory,
            rich_help_panel="Optional",
        ),
    ] = CONFIG.general.output_directory,
    excluded_words: Annotated[
        Path | None,
        typer.Option(
            "--excluded-words",
            "-e",
            callback=CliUtils.validate_optional_file,
            help=HelpText.exclude_sheet,
            rich_help_panel="Optional",
        ),
    ] = CONFIG.external_resources.word_exclude_spreadsheet,
    lemma: Annotated[
        bool,
        typer.Option(
            "--lemma/--no-lemma", "-m/-n", help=HelpText.lemmatise, rich_help_panel="Optional Flags"
        ),
    ] = CONFIG.lemmatiser.lemmatise,
    stop_words: Annotated[
        bool,
        typer.Option(
            "--stop-words/--no-stop-words",
            "-s/-p",
            help=HelpText.stop_words,
            rich_help_panel="Optional Flags",
        ),
    ] = (not CONFIG.lemmatiser.filter_out_stop_words),
):
    CliUtils.print_status("Frequency analysis: starting")

    frequency_analyser = import_module(".frequency_analyser", APP_NAME)
    frequency_analyser.frequency_analyser(
        input_path=input_path,
        output_directory=output_directory,
        language=language,
        lemmatise_words=lemma,
        exclude_spreadsheet_path=excluded_words,
        filter_stop_words=(not stop_words),
    )


# @app.command(
#     no_args_is_help=True,
#     help=HelpText.interactive_transcript,
#     rich_help_panel="Primary Functions",
#     epilog=ffmpeg_warning(),
# )
# def interactive_transcript():
#     # TODO: [V2] Implement
#     return


@app.command(
    no_args_is_help=True,
    rich_help_panel="Primary Functions",
    help=HelpText.transcribe,
    epilog=ffmpeg_warning(),
)
def transcribe(
    input_path: Annotated[
        Path,
        typer.Option(
            "--input",
            "-i",
            help=HelpText.transcribe_path,
            show_default=False,
            callback=CliUtils.validate_path_exists,
            rich_help_panel="Required",
        ),
    ],
    language: Annotated[
        str,
        typer.Option(
            "--language",
            "-l",
            help=HelpText.language_code,
            callback=CliUtils.validate_language_code,
            rich_help_panel="Required",
        ),
    ] = CONFIG.general.language,
    output_directory: Annotated[
        Path,
        typer.Option(
            "--output",
            "-o",
            callback=CliUtils.validate_directory,
            help=HelpText.output_directory,
            rich_help_panel="Optional",
        ),
    ] = CONFIG.general.output_directory,
    whisper_model: Annotated[
        str,
        typer.Option(
            "--whisper-model",
            "-w",
            callback=CliUtils.clean_string,
            help=HelpText.whisper_model,
            rich_help_panel="Optional",
        ),
    ] = CONFIG.transcriber.whisper_model,
    align_model: Annotated[
        str,
        typer.Option(
            "--align-model",
            "-a",
            callback=CliUtils.clean_string,
            help=HelpText.alignment_model,
            rich_help_panel="Optional",
        ),
    ] = CONFIG.transcriber.alignment_model,
    gpu: Annotated[
        bool,
        typer.Option(
            "--gpu/--cpu", "-g/-c", help=HelpText.use_cuda, rich_help_panel="Optional Flags"
        ),
    ] = CONFIG.transcriber.whisper_use_gpu,
    subtitle_format: Annotated[
        str,
        typer.Option(
            "--subtitle-format", "-f", help=HelpText.subtitle_format, rich_help_panel="Optional"
        ),
    ] = CONFIG.transcriber.subtitle_format,
):
    CliUtils.print_status("Transcriber: starting")

    transcriber = import_module(".transcriber", APP_NAME)
    transcriber.transcriber(
        input_path=input_path,
        output_directory=output_directory,
        language=language,
        use_gpu=gpu,
        whisper_model=whisper_model,
        alignment_model=align_model,
        sub_format=subtitle_format,
    )


"""
Configuration commands
"""


@app.command(
    no_args_is_help=True,
    rich_help_panel="Configuration",
    help=HelpText.install,
    epilog=ffmpeg_warning(),
)
def install(
    language: Annotated[
        str,
        typer.Option(
            "--language",
            "-l",
            help=HelpText.language_code,
            callback=CliUtils.validate_language_code,
            rich_help_panel="Required",
        ),
    ] = CONFIG.general.language,
    translation_language: Annotated[
        str,
        typer.Option(
            "--translation-language",
            "-t",
            help=HelpText.language_code_translation,
            callback=CliUtils.validate_language_code,
            rich_help_panel="Optional",
        ),
    ] = CONFIG.general.language_for_translations,
):
    CliUtils.print_status("Updating downloader (yt-dlp)")
    downloader = import_module(".downloader", APP_NAME)
    downloader.downloader_update()

    CliUtils.print_status("Initialiser downloader")
    downloader.downloader_dummy()

    CliUtils.print_status("Checking CUDA status")
    transcriber = import_module(".transcriber", APP_NAME)

    cuda = transcriber.cuda_available()
    if cuda:
        CliUtils.print_rich("CUDA enabled, can use GPU processing")
    else:
        CliUtils.print_rich("CUDA disabled, using CPU processing")

    CliUtils.print_status("Initialising transcriber")
    CliUtils.print_warning(
        "This may appear to freeze for a few minutes if you haven't run it before!"
    )
    dummy_transcribe_file = get_resources_directory() / "a.mp3"
    transcriber.transcriber(
        input_path=dummy_transcribe_file,
        output_directory=get_resources_directory(),
        language=language,
        use_gpu=CONFIG.transcriber.whisper_use_gpu,
        whisper_model=CONFIG.transcriber.whisper_model,
        alignment_model=CONFIG.transcriber.alignment_model,
        sub_format=CONFIG.transcriber.subtitle_format,
    )

    CliUtils.print_status("Initialising lemmatiser")
    lemmatiser = import_module(".lemmatiser", APP_NAME)
    lemmatiser.lemma_dummy(language)

    CliUtils.print_status("Initialising translator")
    translator = import_module(".translator", APP_NAME)
    translator.install(language, translation_language)


@app.command(rich_help_panel="Configuration", help=HelpText.list_languages)
def list_languages(
    detailed: Annotated[
        bool,
        typer.Option(
            "--detailed",
            "-a",
            help=HelpText.list_languages_detailed,
            rich_help_panel="Optional Flags",
        ),
    ] = False,
):
    CliUtils.print_rich(SupportedLanguages.common_explanation)
    CliUtils.print_rich(SupportedLanguages.get_common_languages())

    if detailed:
        CliUtils.print_rich("\n")
        CliUtils.print_rich(SupportedLanguages.whisper_explanation)
        CliUtils.print_rich(SupportedLanguages.whisper_languages)
        CliUtils.print_rich("\n")
        CliUtils.print_rich(SupportedLanguages.spacy_explanation)
        CliUtils.print_rich(SupportedLanguages.spacy_languages)
        CliUtils.print_rich("\n")
        CliUtils.print_rich(SupportedLanguages.argos_explanation)
        CliUtils.print_rich(SupportedLanguages.argos_languages)


@app.command(no_args_is_help=True, rich_help_panel="Configuration", help=HelpText.set_defaults)
def set_defaults(
    factory: Annotated[
        bool | None,
        typer.Option(
            "--factory", "-f", help=HelpText.defaults_factory, rich_help_panel="Optional Flags"
        ),
    ] = None,
    custom: Annotated[
        bool | None,
        typer.Option(
            "--custom", "-c", help=HelpText.defaults_custom, rich_help_panel="Optional Flags"
        ),
    ] = None,
):
    if factory:
        CONFIG.factory_reset()
    elif custom:
        CONFIG.launch_config_file_in_editor()


@app.command(rich_help_panel="Configuration", help=HelpText.update_downloader)
def update_downloader():
    CliUtils.print_status("Downloader update: starting")
    downloader = import_module(".downloader", APP_NAME)
    downloader.downloader_update()


"""
Define callbacks and subcommands.
"""


def version_callback(value: bool):
    """Return version number"""
    if value:
        CliUtils.print_rich(get_version_number())

        CliUtils.print_rich(f"\n{HelpText.license}")

        raise typer.Exit()


@app.callback(help=HelpText.main)
def main(
    version: Annotated[
        bool | None,
        typer.Option("--version", callback=version_callback, help=HelpText.application_version),
    ] = None,
):
    return


if __name__ == "__main__":
    app()



import html
import json
from pathlib import Path

import pandas as pd

from .cli_utils import CliUtils


def generate_output_path(
    source_material_path: Path, output_directory: Path, file_description: str, file_format: str
) -> Path:
    """Create a standardised file path based upon a source media file or directory"""

    file_stem = source_material_path.stem
    video_suffix = source_material_path.suffix.replace(".", "")
    output_path = output_directory / f"{file_stem}_{video_suffix}_{file_description}.{file_format}"

    return output_path


def write_to_json(data: dict | list, file_path: Path) -> None:
    """Write a dict or list to json"""
    try:
        with open(file_path.resolve(), "w") as file:
            json.dump(data, file, indent=4)
    except Exception as e:
        CliUtils.print_error(f"Could not write {str(file_path)}", e)


def load_json(file_path: Path, encoding: str = "utf-8") -> list | dict:
    """Load a json into a list or dict, as appropriate"""
    with open(file_path, "r", encoding=encoding) as file:
        data = json.load(file)

    return data


def dataframe_to_excel(df: pd.DataFrame, file_path: Path, include_index: bool = False) -> None:
    """Save a pandas dataframe to an excel format"""

    try:
        df.to_excel(file_path, index=include_index)
    except Exception as e:
        CliUtils.print_error(f"Could not write {str(file_path)}", e)


def read_file_to_string(file_path: Path, encoding: str = "utf-8") -> str:
    """Read a text based file into memory"""
    output = ""

    with file_path.open("r", encoding=encoding) as file:
        output = file.read()

    return output


def import_first_column_from_sheets(file_path: Path) -> pd.DataFrame:
    """Import the contents of the first column of every sheet in a spreadsheet to dataframe. Ignore the header row."""

    excel_file = pd.ExcelFile(file_path)

    df = pd.DataFrame(columns=["word"])

    # Loop through each sheet
    for sheet_name in excel_file.sheet_names:
        # Read the first column of the sheet, ignoring the header
        new_rows = pd.read_excel(file_path, sheet_name=sheet_name, usecols=[0], names=["word"])

        df = pd.concat([df, new_rows], ignore_index=True)

    return df


def list_files_with_extension(
    input_path: Path,
    valid_suffixes: list,
    file_description_text: str = "",
    search_subdirectories: bool = False,
    print_errors: bool = True,
) -> list[Path]:
    """Gets a list of files (filtered by file extension) from an input path which may be a file or directory."""

    if not input_path.exists():
        if print_errors:
            CliUtils.print_error(
                f"Provided input path does not exist! User provided path: {str(input_path)}"
            )

        return []

    if input_path.is_file():
        if input_path.suffix not in valid_suffixes:
            if print_errors:
                CliUtils.print_error(
                    f"File is not a supported {file_description_text} file! User provided path: {str(input_path)}"
                )
                CliUtils.print_rich(f"Supported {file_description_text} file types:")
                CliUtils.print_rich(valid_suffixes)

        return [input_path]

    elif input_path.is_dir():
        search_pattern = "*"
        if search_subdirectories:
            search_pattern = "**/*"

        glob = input_path.glob(search_pattern)

        path_list = []
        for item in glob:
            if item.is_file() and (item.suffix in valid_suffixes):
                path_list.append(item)

        if len(path_list) == 0:
            if print_errors:
                CliUtils.print_error(
                    f"No {file_description_text} supported files found in directory! User provided path: {str(input_path)}."
                )
                CliUtils.print_rich(f"Supported {file_description_text} file types:")
                CliUtils.print_rich(valid_suffixes)

        return path_list
    else:
        if print_errors:
            CliUtils.print_error("Could not get list of files")
        return []


def sanitise_string_html(input_string: str) -> str:
    """Make a string html safe"""

    if isinstance(input_string, str):
        output_string = html.escape(input_string)
        return output_string
    else:
        return ""



import shutil
from pathlib import Path

import pandas as pd
from rich.progress import track

from .config import SUPPORTED_WORD_AUDIO_EXTS
from .utils import list_files_with_extension, load_json


def lookup_all_words(
    df: pd.DataFrame, dictionary_path: Path, source_column: str = "word"
) -> pd.DataFrame:
    """Loops through a dataframe and fetches the definitions for all words"""

    # Get dictionary entries
    entries = load_dictionary(dictionary_path)
    assert isinstance(entries, list)

    # Add the definitions for each of the words
    for index, row in track(df.iterrows(), description="Looking up definitions...", total=len(df)):
        word = str(row[source_column]).lower()

        definition = find_definition(entries=entries, search_term=word)

        df.at[index, "definition"] = definition

    return df


def copy_all_word_audio(
    df: pd.DataFrame, library_path: Path, output_directory: Path, source_column: str = "word"
) -> pd.DataFrame:
    """Loops through a dataframe, searches for audio that matches the word and, if found, copy it to a specific directory.
    Path of copied file is written to the dataframe"""

    # Get list of files
    library = get_sound_library(library_path=library_path)

    # Copy the files to the working directory and add a link
    for index, row in track(df.iterrows(), description="Getting word audio...", total=len(df)):
        word = str(row[source_column])
        word_path = copy_sound(
            sound_library=library, lookup_value=word, output_directory=output_directory
        )

        if word_path:
            df.at[index, "word_audio_path"] = str(word_path.resolve())

    return df


def load_dictionary(dictionary_path: Path) -> list | dict:
    """Load a json dictionary from a specified path"""
    entries = load_json(dictionary_path)

    return entries


def get_sound_library(
    library_path: Path,
    file_extensions: list = SUPPORTED_WORD_AUDIO_EXTS,
    file_decription="word sound files",
) -> dict[str, Path]:
    """Find sound files in a directory and save a dictionary in the format {lower case version of the file stem: full path}"""
    sounds = list_files_with_extension(
        input_path=library_path,
        valid_suffixes=file_extensions,
        file_description_text=file_decription,
        search_subdirectories=True,
    )

    sound_library: dict[str, Path] = {}
    for sound in sounds:
        word = sound.stem.lower().strip().rstrip(".").strip()
        if word not in sound_library:
            sound_library[word] = sound

    return sound_library


def copy_sound(
    sound_library: dict[str, Path], lookup_value: str, output_directory: Path
) -> Path | None:
    """If a sound is found within the library, copy it to a specific directory. If successful, return the path of the copied file."""
    lookup_value = lookup_value.lower()

    if lookup_value not in sound_library:
        return None

    src = sound_library[lookup_value]
    filename = src.name
    dst = output_directory / filename

    shutil.copy(src, dst)

    return dst


def find_definition(entries: list, search_term: str) -> str:
    """Lookup the definition of a word within a json dictionary that has been loaded into memory"""

    if search_term:
        search_term = search_term.lower()

        left = 0
        right = len(entries) - 1

        while left <= right:
            mid = (left + right) // 2
            mid_term = entries[mid]["term"]

            if mid_term == search_term:
                return entries[mid]["definition"]
            elif mid_term < search_term:
                left = mid + 1
            else:
                right = mid - 1

        return ""

    return ""



from pathlib import Path

import yt_dlp

from .cli_utils import CliUtils
from .command_runner import arg_string_to_list, update_package
from .config import APP_NAME
from .ytdlp_cli_to_api import cli_to_api


def downloader(
    url: str,
    download_directory: Path = Path(""),
    format: str = "",
    subtitle_language: str = "",
    extract_audio: bool = False,
    other_options: str = "",
    subs_only: bool = False,
) -> list[Path]:
    """Main entry point for downloading videos / audio using yt-dlp"""

    # Check if subtitles are available
    subs_available = False
    if subtitle_language != "":
        subs_available = check_subs_available(url, subtitle_language)

    if (not subs_available) and subs_only:
        if subtitle_language.strip() == "":
            CliUtils.print_error(
                "Exiting, no subtitles available for download as value of `--subtitle-language` was blank. You can use [cyan bold]gogadget transcribe[/] to generate your own subtitles."
            )
        else:
            CliUtils.print_error(
                f"Exiting, no subtitles available for download for chosen language `{subtitle_language}`. You can use [cyan bold]gogadget transcribe[/] to generate your own subtitles."
            )
        return []

    # Work out the command line arguments that we want to use
    option_args = []

    if download_directory.name != "":
        option_args += ["-P", download_directory.name]
    if format != "":
        option_args += ["-f", format]
    if extract_audio:
        option_args += ["-x", "--audio-format", "mp3"]
    if subtitle_language != "" and subs_available:
        option_args += ["--write-subs", "--sub-lang", subtitle_language, "--write-auto-subs"]
    if subs_only:
        option_args += ["--skip-download"]
    if other_options != "":
        option_args += arg_string_to_list(other_options)

    # Convert the options to api style
    ydl_opts = cli_to_api(option_args, cli_defaults=True)

    # Get list of videos to download (e.g. if it's a playlist)
    CliUtils.print_status("Downloader: getting list of videos to download")
    url_list = url_to_list(url=url, ydl_opts=ydl_opts)

    # Extract the video(s)
    CliUtils.print_status("Downloader: downloading files")
    file_list = download_videos(urls=url_list, ydl_opts=ydl_opts)

    return file_list


def download_videos(urls: list[str], ydl_opts: dict) -> list[Path]:
    """Loop through urls and download each of them"""
    try:
        with yt_dlp.YoutubeDL(ydl_opts) as ydl:
            file_list = []
            for url in urls:
                info_dict = ydl.sanitize_info(ydl.extract_info(url, download=True))
                file_path = Path(ydl.prepare_filename(info_dict))
                file_list.append(file_path)

            return file_list

    except Exception as e:
        CliUtils.print_error("Could not download videos", e)
        return []


def url_to_list(url: str, ydl_opts: dict) -> list[str]:
    """Unwrap playlists so that they are treated as individual items in the list"""

    try:
        with yt_dlp.YoutubeDL(ydl_opts) as ydl:
            info_dict = ydl.sanitize_info(ydl.extract_info(url, download=False))
            link_type = info_dict.get("_type", "video")

            urls = []
            if link_type == "video":
                urls.append(url)
            elif link_type == "playlist":
                urls = []
                if "entries" in info_dict:
                    for entry in info_dict["entries"]:
                        urls.append(entry["url"])
            else:
                CliUtils.print_warning(
                    f"Downloader does not currently support url of type {link_type}. May not function correctly."
                )

            return urls
    except Exception as e:
        CliUtils.print_error("Could not parse download urls", e)
        return []


def check_subs_available(url: str, language: str) -> bool:
    """Determine if subtitles are available on the website for a specified language code.
    Note that some websites don't use codes fully compliant with ISO (e.g. the might do en_to_it rather than just it)
    """

    try:
        option_args = ["--write-subs", "--sub-lang", language, "--write-auto-subs"]
        ydl_opts = cli_to_api(option_args, cli_defaults=False)
        ydl_opts["skip_downloads"] = True

        with yt_dlp.YoutubeDL(ydl_opts) as ydl:
            info_dict = ydl.sanitize_info(ydl.extract_info(url, download=False))

            captions_list: list[str] = []
            if "automatic_captions" in info_dict:
                captions_list = captions_list + list(info_dict["automatic_captions"].keys())

            if "subtitles" in info_dict:
                captions_list = captions_list + list(info_dict["subtitles"].keys())

            if len(captions_list) == 0:
                CliUtils.print_warning("No captions available for download for this video")
                return False
            elif language not in captions_list:
                CliUtils.print_warning(
                    f"No captions available in language {language} for this video"
                )
                return False

        return True
    except Exception:
        CliUtils.print_warning("Could not extract captions information for this video")
        return False


def downloader_dummy() -> None:
    """Dummy object to initialise yt-dlp on the very first run of the tool"""
    yt_dlp.YoutubeDL

    return


def downloader_update() -> None:
    """Update the yt-dlp python package"""
    update_package("yt-dlp", APP_NAME)



from pathlib import Path

from .cli_utils import CliUtils
from .command_runner import run_command
from .config import SUPPORTED_AUDIO_EXTS, SUPPORTED_VIDEO_EXTS
from .utils import list_files_with_extension


def transcriber(
    input_path: Path,
    output_directory: Path,
    language: str,
    use_gpu: bool,
    whisper_model: str,
    alignment_model: str,
    sub_format: str,
) -> list:
    """Main entry point for the media file transcriber"""

    # Get media files in path (path could be a file or a directory)
    supported_formats = SUPPORTED_VIDEO_EXTS + SUPPORTED_AUDIO_EXTS
    path_list = list_files_with_extension(
        input_path,
        valid_suffixes=(SUPPORTED_VIDEO_EXTS + SUPPORTED_AUDIO_EXTS),
        file_description_text="media files",
    )

    if len(path_list) == 0:
        CliUtils.print_warning("No supported file formats found")
        CliUtils.print_rich("Supported formats:")
        CliUtils.print_rich(supported_formats)
        return []

    # Configure settings
    output_dir_str = str(output_directory.resolve())
    compute_type = "int8"
    device = "cpu"
    if use_gpu:
        if cuda_available():
            device = "cuda"
            compute_type = "float16"
        else:
            CliUtils.print_warning(
                """You have requested --gpu but CUDA is not configured.
Troubleshooting:
    - If you are on windows, did you check the CUDA option in the installer?
    - Please see readme for more information"""
            )
            CliUtils.print_warning("Falling back to --cpu")

    # Run for each file
    for file_path in path_list:
        file_str = str(file_path.resolve())

        command = [
            "whispergg",
            file_str,
            "--compute_type",
            compute_type,
            "--device",
            device,
            "--language",
            language,
            "--model",
            whisper_model,
            "--print_progress",
            "True",
            "--output_format",
            sub_format,
            "--output_dir",
            output_dir_str,
        ]

        if alignment_model != "":
            command += ["--align_model", alignment_model]

        run_command(command, print_command=True)

    return path_list


def cuda_available() -> bool:
    from torch.cuda import is_available

    return is_available()


# def configure_cuda() -> None:
#     if not cuda_available():
#         install_package(
#             package_name="torch==2.5.1+cu124",
#             package_index="https://download.pytorch.org/whl/cu124",
#             app_name=APP_NAME,
#         )
#         install_package(
#             package_name="torchaudio==2.5.1+cu124",
#             package_index="https://download.pytorch.org/whl/cu124",
#             app_name=APP_NAME,
#         )
